/*-------------------------------------------------------------------
|                             SHOWWPG.C                              |
--------------------------------------------------------------------*/
#include <windows.h>
#include <string.h>      /* _fstrrchr */
#include <math.h>        /* cos, sin */
//#include <string.h>         /* memset */
#include <commdlg.h>     /* GetOpenFileName() */
#include "showwpg.h"
#include "globals.h"

#ifdef __BORLANDC__
    #pragma argsused
#endif
int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                   LPSTR lpszCmdParam, int nCmdShow)
{
  static char szMenuName[] = "SWPGMenu";
  HWND     hWnd;
  MSG      msg;
  WNDCLASS wndclass;
  HANDLE   hAccel;

  /* Grab hInstance into global for later use. */
  hInst = hInstance;

  if (!hPrevInstance)
    {
    wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
    wndclass.lpfnWndProc   = WndProc;
    wndclass.cbClsExtra    = 0;
    wndclass.cbWndExtra    = 0;
    wndclass.hInstance     = hInstance;
    wndclass.hIcon         = LoadIcon(hInstance, "SWPGIcon");
    wndclass.hCursor       = NULL;
    wndclass.hbrBackground = GetStockObject(WHITE_BRUSH);
    wndclass.lpszMenuName  = szMenuName;
    wndclass.lpszClassName = AppName;
    RegisterClass(&wndclass);
    }  /* if */

  /* Load global cursor handles. */
  HCursor[TOP_LEFT] = HCursor[BOT_RT]   = LoadCursor((HINSTANCE)NULL, IDC_SIZENWSE);
  HCursor[TOP_MID]  = HCursor[BOT_MID]  = LoadCursor((HINSTANCE)NULL, IDC_SIZENS);
  HCursor[TOP_RT]   = HCursor[BOT_LEFT] = LoadCursor((HINSTANCE)NULL, IDC_SIZENESW);
  HCursor[MID_RT]   = HCursor[MID_LEFT] = LoadCursor((HINSTANCE)NULL, IDC_SIZEWE);
  HCursor[TRANSLATE] = LoadCursor((HINSTANCE)NULL, IDC_SIZE);
  HCursor[NO_HIT] = LoadCursor((HINSTANCE)NULL, IDC_ARROW);
  HCursor[WAIT] = LoadCursor((HINSTANCE)NULL, IDC_WAIT);

  hWnd = CreateWindow(AppName, "ShowWPG", WS_OVERLAPPEDWINDOW,
                      INIT_POSX, INIT_POSY, INIT_WIDTH, INIT_HEIGHT,
                      (HWND)NULL, (HMENU)NULL, hInstance, NULL);

  ShowWindow(hWnd, nCmdShow);
  UpdateWindow(hWnd);

  /* Gray 'Copy' menu item; load keyboard accelerators. */
  EnableMenuItem(GetMenu(hWnd), IDM_COPY, MF_GRAYED);
  hAccel = LoadAccelerators(hInstance, AppName);

  while (GetMessage(&msg, (HWND)NULL, 0, 0))
    {
    if (!TranslateAccelerator(hWnd, hAccel, &msg))
      {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
      }  /* if */
    }  /* while */

  return msg.wParam;
}

long WINAPI _export WndProc(HWND hWnd, UINT message, UINT wParam, LONG lParam)
/*
   USE:  Main window procedure.
   IN:   Standard WndProc() parameters.
   NOTE: WM_CREATE     : Put AppName - (Untitled) in caption.
         WM_ACTIVATE   : To set AppName only as caption when minimized.
         WM_SIZE       : Update global ClientRect.
         WM_COMMAND    : IDM_RETRIEVE triggers GetOpenFileName()
                         IDM_COPY copies to clipboard as MetaFile
                         IDM_ABOUT shows About DlgBox
         WM_PAINT      : Traverse and display WPG file in TargetRect
         WM_MOUSEMOVE  : Pass along to Modify() to size or translate
         WM_LBUTTONDOWN:   "
         WM_LBUTTONUP  :   "
         WM_LBUTTONDBLCLK:  Pass along to Modify() to zoom in
         WM_RBUTTONDBLCLK:  Pass along to Modify() to zoom out
         WM_DESTROY    : Finalize.

         hBuf is a key global variable that is non-NULL if and only if
         it is a valid memory handle to a WPG file. Check hBuf to see
         if you have a file to show.
*/
{
  HANDLE hNewMem;   /* temp handle to buffer for file */
  PAINTSTRUCT ps;   /* needed in WM_PAINT */
  HDC    hDC;       /* needed in WM_PAINT */
  LPSTR  lpls;      /* ptr to last '\' in FileName string */
  static FARPROC lpDlg;   /* for About box function */

  switch (message)
    {
    case WM_CREATE:
      /* Write default caption for main window. */
      wsprintf(Caption, "%s - (Untitled)", (LPSTR) AppName);
      SetWindowText(hWnd, Caption);
      /* Default iconic name is just app name. */
      lstrcpy(IconicName, AppName);
      break;  /* WM_CREATE */

    case WM_ACTIVATE:
      /* When minimized, use AppName only for caption, else full caption. */
      SetWindowText(hWnd, IsIconic(hWnd) ? IconicName : Caption);
      break;  /* WM_ACTIVATE */

    case WM_SIZE:
      /* Update global ClientRect. */
      GetClientRect(hWnd, &ClientRect);
      break;  /* WM_SIZE */

    case WM_COMMAND:
      switch (wParam)
        {
        case IDM_RETRIEVE:
          /* Get file name, read into memory; break if can't. */
          if ((hNewMem = LoadWPGFile(hWnd)) == NULL)
            break;
          /* If hBuf valid, free it; set hBuf to new handle. */
          if (hBuf) GlobalFree(hBuf);
          hBuf = hNewMem;
          break;  /* IDM_RETRIEVE */

        case IDM_COPY:
          /* hBuf is non-NULL means file is in memory. */
          if (hBuf)  CopyWMFToClipBoard(hWnd);   
          break;  /* IDM_COPY */

        case IDM_ABOUT:
          lpDlg = MakeProcInstance((FARPROC) AboutDlgProc, hInst);
          DialogBox(hInst, "AboutBox", hWnd, (DLGPROC)lpDlg);
          FreeProcInstance(lpDlg);
          break;  /* IDM_ABOUT */
        }
      break;  /* WM_COMMAND */

    case WM_PAINT:
      hDC = BeginPaint(hWnd, &ps);
      if (hBuf)  PaintWindow(hDC);
      EndPaint(hWnd, &ps);
      break;  /* WM_PAINT */

    case WM_MOUSEMOVE:
    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_LBUTTONDBLCLK:
    case WM_RBUTTONDBLCLK:
      if (hBuf)
        Modify(hWnd, message, lParam);
      else
        /* SetCursor() on every MOUSEMOVE when no file loaded. */
        SetCursor(HCursor[NO_HIT]);
      break;  /* WM_MOUSEMOVE... */

    case WM_DESTROY:
      /* Free memory handle if valid. */
      if (hBuf)  GlobalFree(hBuf);
      if ((lpls = _fstrrchr(FileName, '\\')) != NULL)
        {
        /* Excise file name from FileName string so only path left. */
        *lpls = 0;
        /* Write current path string into SHOWWPG.INI. */
        WritePrivateProfileString("Defaults", "Path", FileName, "SHOWWPG.INI");
        }
      PostQuitMessage(0);
      break;

    default:  /* Passes it on if unproccessed */
      return (DefWindowProc(hWnd, message, wParam, lParam));
    }  /* switch */

  return (NULL);
}

HANDLE LOCAL LoadWPGFile(HWND hWnd)
/*
   USE:  Load WPG file from into memory block.
   IN:   hWnd = main hWnd for common dialog box, menu work, etc.
   RET:  handle to memory block containing file or NULL if not loaded
   NOTE: The function is successful when a new file has been loaded;
         in this case the function returns the memory handle containing
         the file, which the caller must later free. Several global
         variables are updated accordingly:

               FileName:   current WPG file
               Caption:    window caption string
               IconicName: caption string when iconized
               Scale:      current scale factors and offsets
               TargetRect: current target rectangle for image
*/
{
  int    ret;          /* return value from FileRead() */
  HANDLE hWPGMem;      /* temp handle to buffer for file */
  char   szError[20];  /* buffer for error string */
  LPSTR  lpls;         /* ptr to last '\' in FileName string */
  LPSTR  lpper;        /* ptr to '.' in IconicName string */

  /* Call dialog box to get file name from user; return NULL if can't. */
  if (!OpenDlgBox(hWnd, FileName, sizeof(FileName)))
    return (NULL);

  /* Load memory handle with file; return NULL if can't */
  if ((ret = FileRead(FileName, &hWPGMem)) != SUCCESS)
    {
    /* Load resource string in RC with error string, return NULL. */
    LoadString(hInst, IDS_ERROR_BASE + ret, szError, sizeof(szError));
    MessageBox(hWnd, szError, AppName, MB_OK | MB_ICONEXCLAMATION);
    return (NULL);
    }

  /* If got thru to here, hWPGMem has been successfully loaded. */

  /* Set Scale parameters and TargetRect. */
  InitScaleParms(hWPGMem, &Scale, &TargetRect);

  /* Enable 'Copy' menu item. */
  EnableMenuItem(GetMenu(hWnd), IDM_COPY, MF_ENABLED);
  /* Rewrite caption to include path + file name. */
  wsprintf(Caption, "%s - %s", (LPSTR) AppName, (LPSTR) FileName);
  SetWindowText(hWnd, Caption);

  /* Reload iconic name to include file name (not path or ext). */
  if ((lpls = _fstrrchr(FileName, '\\')) != NULL)
    {
    /* Position just past '\' to file name. */
    lpls++;
    wsprintf(IconicName, "%s %s", (LPSTR) AppName, (LPSTR) lpls);
    /* Chop extension if found (probably ".WPG"). */
    if ((lpper = _fstrrchr(IconicName, '.')) != NULL)
      *lpper = 0;
    }

  /* Force WM_PAINT to put image on screen, return handle to memory. */
  InvalidateRect(hWnd, NULL, TRUE);
  return (hWPGMem);
}

#ifdef __BORLANDC__
    #pragma argsused
#endif
BOOL WINAPI _export AboutDlgProc(HWND hDlg, UINT message, UINT wParam, LONG lParam)
/*
   USE:  WndProc() for About Box.
   IN:   Standard dialog box parameters.
*/
{
  switch (message)
    {
    case WM_INITDIALOG:
      return TRUE;
    case WM_COMMAND:
      switch (wParam)
        {
        case IDOK:
        case IDCANCEL:
          EndDialog(hDlg, 0);
          return(TRUE);
         }
      break;
    }
  return(FALSE);
}

void LOCAL PaintWindow(HDC hDC)
/*
   USE:  Display WPG file in window when receive WM_PAINT.
   IN:   hDC = app's DC from BeginPaint()
   NOTE: Use clip regions so image kept off instruction line at bottom.
*/
{
  HCURSOR hOldCur; /* to restore original cursor after wait cursor */
  HRGN hRgn;       /* to set clip region excluding instruction line */
  HRGN hOldRgn;    /* to reselect original clip region */

  hOldCur = SetCursor(HCursor[WAIT]);
  PaintInstructionLine(hDC);
  hRgn = CreateRectRgn(0, 0, ClientRect.right, ClientRect.bottom-(ILINE_HT+2));
  hOldRgn = SelectObject(hDC, hRgn);
  FrameOurRect(hDC, &TargetRect);
  PlayWPGFile(hBuf, hDC, Scale);
  SelectObject(hDC, hOldRgn);
  DeleteObject(hRgn);
  SetCursor(hOldCur);
}

void LOCAL Modify(HWND hWnd, UINT message, LONG lParam)
/*
   USE:  Sizing and translating tool.
   IN:   WndProc() parameters passed along.
   NOTE: Simple state tool with route WAITING --> START_MODIFYING -->
         MODIFYING --> WAITING. START_MODIFYING is the state for the first
         WM_MOUSEMOVE after the tool has started. There are 9 modifications:
         translation and 8 types of resizing. Which modification is
         triggered depends on which hot spot the cursor was in at
         WM_LBUTTONDOWN. Each modification has its own section in
         RestrictCursor(), DrawDraggingRect(), and UpdateScale(),
         functions called respectively at WM_LBUTTONDOWN, WM_MOUSEMOVE,
         and WM_LBUTTONUP (RestrictCusor initializes, DrawDraggingRect
         draws interactively, UpdateScale finalizes).
*/
{
  HDC hDC;           /* must generate our own handle to DC to draw */
  static POINT pt;   /* button-down point */
  static HPEN hPen;  /* dashed pen */
  static HRGN hRgn;  /* clip region for client rect excepting instr line */
  static int hs;     /* hot spot index */
  static int state;  /* system state */

  switch (message)
    {
    case WM_LBUTTONDOWN:
      if (state == WAITING)
        if ((hs = PtinHotSpot(&TargetRect, MAKEPOINT(lParam))) != NO_HIT)
          {
          state = START_MODIFYING;
          hPen = CreatePen(PS_DOT, 1, 0L);
          hRgn = CreateRectRgn(0,0,ClientRect.right,ClientRect.bottom-(ILINE_HT+2));
          pt = MAKEPOINT(lParam);
          RestrictCursor(hs, hWnd, &pt);
          }
      break;  /* WM_LBUTTONDOWN */

    case WM_MOUSEMOVE:
      switch (state)
        {
        case WAITING:
          /* Set Cursor appropriately if in hot spot, Arrow if not. */
          SetCursor(HCursor[PtinHotSpot(&TargetRect, MAKEPOINT(lParam))]);
          break;
        case START_MODIFYING:
          state = MODIFYING;
          hDC = GetDC(hWnd);
          SelectObject(hDC, hRgn);
          /* Erase Framed rect (XOR mode). */
          SetROP2(hDC, R2_NOTXORPEN);
          FrameOurRect(hDC, &TargetRect);
          /* Draw rect with SOLID pen first time. */
          DrawDraggingRect(hs, hDC, pt);
          pt = MAKEPOINT(lParam);
          /* Switch to DASHED pen for next draw rect. */
          SelectObject(hDC, hPen);
          DrawDraggingRect(hs, hDC, pt);
          ReleaseDC(hWnd, hDC);
          break;
        case MODIFYING:
          hDC = GetDC(hWnd);
          /* All drawing in XOR mode with DASHED pen, clipped to region. */
          SetROP2(hDC, R2_NOTXORPEN);
          SelectObject(hDC, hPen);
          SelectObject(hDC, hRgn);
          /* Erase last rect. */
          DrawDraggingRect(hs, hDC, pt);
          pt = MAKEPOINT(lParam);
          /* Draw new rect. */
          DrawDraggingRect(hs, hDC, pt);
          ReleaseDC(hWnd, hDC);
          break;
        }  /*  switch (state) */
      break;  /* WM_MOUSEMOVE */

    case WM_LBUTTONUP:
      if (state == MODIFYING || state == START_MODIFYING)
        {
        DeleteObject(hPen);
        DeleteObject(hRgn);
        UpdateScale(hs, pt);
        /* Set Cursor appropriately if in hot spot, Arrow if not. */
        SetCursor(HCursor[PtinHotSpot(&TargetRect, MAKEPOINT(lParam))]);
        /* RePaint only if modification started. */
        if (state == MODIFYING) InvalidateRect(hWnd, NULL, TRUE);
        state = WAITING;
        }
      break;  /* WM_LBUTTONUP */

    case WM_LBUTTONDBLCLK:
      if (state == WAITING && PtInRect(&TargetRect, MAKEPOINT(lParam)))
        {
        UpdateScale(ZOOM_IN, pt);
        InvalidateRect(hWnd, NULL, TRUE);
        }
      break;  /* WM_LBUTTONDBLCLK */

    case WM_RBUTTONDBLCLK:
      if (state == WAITING && PtInRect(&TargetRect, MAKEPOINT(lParam)))
        {
        UpdateScale(ZOOM_OUT, pt);
        InvalidateRect(hWnd, NULL, TRUE);
        }
      break;  /* WM_RBUTTONDBLCLK */
    }  /* switch (message) */
}

long LOCAL FileLength(int hFile)
/*
   USE:  Return number of bytes in file.
   IN:   File handle returned by _lopen().
   NOTE: See Petzold, Programming Windows, Chap 10 (POPFILE.C)
*/
{
  long lCurrentPos = _llseek(hFile, 0L, 1);
  long lFileLength = _llseek(hFile, 0L, 2);
  
  _llseek(hFile, lCurrentPos, 0);
  return(lFileLength);
}

int LOCAL FileRead(LPSTR FileName, LPHANDLE lphMem)
/*
   USE:  Read file from disk.
   IN:   FileName = DOS file name.
   OUT:  *lphMem = memory handle containing file
   RET:  Success or failure status (SUCCESS, FILE_OPEN_FAILURE, etc.).
   NOTE: If successful, memory handle passed back to caller. Caller
         must GlobalFree() handle later if got valid one.
*/
{
  int    hFile;      /* handle to file to open */
  long   lLength;    /* file length, in bytes */
  LPBYTE lpBuf;      /* ptr to buffer */

  /* Initialize as NULL until GlobalAlloc() assigns it. */
  *lphMem = NULL;

  /* Open file for read. */
  if (-1 == (hFile = _lopen(FileName, OF_READ | OF_SHARE_DENY_WRITE)))
    return (FILE_OPEN_FAILURE);

  /* Get file length; return FILE_TOO_LARGE if too large. */
  if ((lLength = FileLength(hFile)) >= MAX_FILE_SIZE)
    {
    _lclose(hFile);
    return (FILE_TOO_LARGE);
    }

  /* GlobalAlloc() for entire file. */
  if (NULL == (*lphMem = GlobalAlloc(GHND, lLength)))
    {
    _lclose(hFile);
    return (NOT_ENOUGH_MEMORY);
    }

  /* Lock mem handle and read file into it. */
  lpBuf = GlobalLock(*lphMem);
  if (_lread(hFile, lpBuf, (WORD)lLength) != (UINT) lLength)
    {
    _lclose(hFile);
    /* Unlock and free mem handle. */
    GlobalUnlock(*lphMem);
    GlobalFree(*lphMem);
    *lphMem = NULL;
    return (FILE_READ_FAILURE);
    }

  /* Close file. */
  _lclose(hFile);

  /* Check signature bytes in prefix to verify WPG file. */
  if ((*lpBuf != 0xFF) || (*(lpBuf+1) != 'W') || (*(lpBuf+2) != 'P') ||
      (*(lpBuf+3) != 'C') || (*(lpBuf+9) != 0x16))
    {
    MessageBeep(0);
    /* Unlock and free mem handle. */
    GlobalUnlock(*lphMem);
    GlobalFree(*lphMem);
    *lphMem = NULL;
    return (NOT_WPG_FILE);
    }

  /* Caller gets valid memory handle; unlock it and return SUCCESS. */
  GlobalUnlock(*lphMem);
  return (SUCCESS);
}

BYTE LOCAL DrawWPGRecord(HDC hDC, LPBYTE FAR *p, SCALE sc)
/*
   USE:  Decipher and draw WPG record.
   IN:   hDC  = handle to DC in which to draw
         *p = ptr to start of WPG record
         sc = SCALE struct to use in drawing record
   OUT:  *p = ptr to start of next record
   RET:  Return record type.
   NOTE: p is a ptr to a ptr to the start of the WPG record. The ptr
         scans thru the record and ends by pointing to the next
         record. The actual ptr, maintained by the caller, is updated
         as well (this is why we need a ptr to a ptr). Note that:

                        p = ptr to ptr
                       *p = ptr
                      **p = data
*/
{
  BYTE  recType;     /* WPG record type */
  DWORD recLen;      /* WPG record length, as DWORD */

  /* Get RecordType, advance ptr. */
  recType  = *(*p)++;

  recLen = GetRecordLength(p);

  /* Process record. Each type of record being handled appears as a 
   * case constant; records not listed are ignored and jumped over.
   */
  switch(recType)
    {
    case WPG_FILLATTR:
      SetBrush(hDC, *p, SET_MID);
      break;
    case WPG_LINEATTR:
      SetPen(hDC, *p, SET_MID);
      break;
    case WPG_LINE:
      DrawLine(hDC, *p, sc);
      break;
    case WPG_POLYLINE:
    case WPG_POLYGON:
      DrawPoly(hDC, *p, sc, recType);
      break;
    case WPG_ELLIPSE:
      DrawEllipse(hDC, *p, sc);
      break;
    }  /* switch */

  /* Advance ptr to next record for caller. */
  *p += recLen;

  return(recType);
}

DWORD LOCAL GetRecordLength(LPBYTE FAR *p)
/*
   USE:  Get record length from WPG record.
   IN:   *p = ptr to start of record length bytes
   OUT:  *p = ptr to record data (just past record length bytes)
   RET:  record length as DWORD
*/
{
  BYTE  byteLen;     /* WPG record length (1st BYTE) */
  DWORD recLen;      /* WPG record length, as DWORD */
  WORD  wordLen;     /* WPG record length (1st WORD) */

  /* Get 1st BYTE of RecordLength, advance ptr. */
  byteLen  = *(*p)++;

  /* Length encoded in BYTE -- return as DWORD. */
  if (byteLen != 0xFF)
    return ((DWORD)byteLen);

  /* If get to here, length encoded in WORD or DWORD. */

  /* Get 1st WORD of RecordLength. */
  wordLen = *((LPWORD)*p);
  /* Length encoded in WORD -- advance ptr and return as DWORD. */
  if (!(wordLen & 0x8000))
    {
    *p += sizeof(WORD);
    return ((DWORD)wordLen);
    }

  /* If get to here, length encoded in DWORD. */

  wordLen &= 0x7FFF;    
  *p += sizeof(WORD);
  /* Put hi part (wordLen), lo part (**p) together by shifting. */
  recLen = ((DWORD)wordLen << 16) + *((LPWORD)*p);
  /* Advance ptr to data. */
  *p += sizeof(WORD);
  return (recLen);
}

void LOCAL SetBrush(HDC hDC, LPBYTE q, int when)
/*
   USE:  Create and select brush.
   IN:   hDC = DC in which to select brush.
         q = ptr to WPG FILLATTR data
         when = (SET_INIT, SET_MID, SET_END) for 1st, middle, or last call
   NOTE: WPG_FILLATTR byte#1 = fill pattern (65 possible)
                      byte#2 = fill color
         Must delete previously created brushes. Calling logic must call
         here once at the start with SET_INIT, then possibly multiple times
         with SET_MID (to make settings from WPG_FILLATTR records), then once
         at end with SET_END; all to make sure previously created brushes are
         deleted (after selecting a new brush into the DC).
*/
{
  BYTE   pattern;     /* fill pattern (64 possible) */
  BYTE   color;       /* fill color */
  HBRUSH hBrush;      /* brush to be created and selected */
  LOGBRUSH logBrush;  /* to create hollow brush */
  static HBRUSH hLastBrush;   /* last brush selected -- deleted later */

  switch (when)
    {
    case SET_INIT:
      hLastBrush = CreateSolidBrush(RGB(0xFF, 0xFF, 0xFF));
      SelectObject(hDC, hLastBrush);
      break;
    case SET_MID:
      /* Get fill pattern and fill color from WPG record. */
      pattern = *q++;
      color   = *q;
      /* Create brush needed according to WPG info. */
      if (pattern == FA_HOLLOW)
        {
        /* lbStyle is only field used for hollow brush. */
        logBrush.lbStyle = BS_HOLLOW;
        hBrush = CreateBrushIndirect(&logBrush);
        }
      else if (pattern == FA_SOLID)
        hBrush = CreateSolidBrush(WPGtoWINColor[color]);
      else if (pattern >= FA_NARROW45 && pattern <= FA_UMBRELLA2)
        hBrush = MakePatternBrush(FillPattern[pattern], WPGtoBMPColor[color]);
      else if (pattern == FA_BACKGROUND)
        /* Supposed to be background color, but WP iself uses WHITE. */
        hBrush = CreateSolidBrush(WPGtoWINColor[15]);
      else
        /* Shouldn't get here, but create brush for symmetry in case you do.*/
        hLastBrush = CreateSolidBrush(RGB(0xFF, 0xFF, 0xFF));
      /* Select created brush, delete last one, set static to new brush. */
      SelectObject(hDC, hBrush);
      DeleteObject(hLastBrush);
      hLastBrush = hBrush;
      break;  /* case SET_MID */
    case SET_END:
      SelectObject(hDC, GetStockObject(WHITE_BRUSH));
      DeleteObject(hLastBrush);
      break;
    }  /* switch */
}

HBRUSH LOCAL MakePatternBrush(LPBYTE mp, int color)
/*
   USE:  Makes 8x8 color bitmap from monochrome template.
   IN:   mp = 8 BYTE monochrome pattern -- template for 8x8 bitmap
         color = WPG color to use in making new bitmap (0 - 15)
   RET:  handle to color pattern brush based on monochrome template
   NOTE: Important not to alter mp[] so subsequent calls can use template.
         WPG color must be converted to BMP color index before call.
*/
{
  HGLOBAL hbmMem;    /* mem handle for our bitmap */
  LPBYTE  bmp;       /* ptr to mem block for DIB */
  LPDWORD dp;        /* ptr to load ColorTable */
  LPBYTE  p;         /* ptr to load bitmap bits */
  int     i, j;      /* to traverse bits of mp[] */
  BYTE    row;       /* to equal mp[i] -- this copy altered */
  WORD    hi;        /* hi bit of row == mp[i] */
  HBRUSH  hBrush;    /* pattern brush to return */

  /* 16 = #entries in ColorTable; 32 = ImageSize. */
  hbmMem = GlobalAlloc(GHND, sizeof(BITMAPINFOHEADER)+16*sizeof(DWORD)+32);
  bmp = (LPBYTE) GlobalLock(hbmMem);

  /* Set BitmapInfoHeader (40 bytes). */
  ((LPBITMAPINFOHEADER)bmp)->biSize = 40;
  ((LPBITMAPINFOHEADER)bmp)->biWidth = 8;
  ((LPBITMAPINFOHEADER)bmp)->biHeight = 8;
  ((LPBITMAPINFOHEADER)bmp)->biPlanes = 1;
  ((LPBITMAPINFOHEADER)bmp)->biBitCount = 4;
  ((LPBITMAPINFOHEADER)bmp)->biCompression = BI_RGB;
  ((LPBITMAPINFOHEADER)bmp)->biSizeImage = 32;
  ((LPBITMAPINFOHEADER)bmp)->biXPelsPerMeter = 0;
  ((LPBITMAPINFOHEADER)bmp)->biYPelsPerMeter = 0;
  ((LPBITMAPINFOHEADER)bmp)->biClrUsed = 0;
  ((LPBITMAPINFOHEADER)bmp)->biClrImportant = 0;

  /* Set DWORD ptr dp just past bmih. */
  dp = (LPDWORD) (bmp + sizeof(BITMAPINFOHEADER));

  /* Load ColorTable with standard Windows colors. */
  *dp++ = RGB(  0,   0,   0);
  *dp++ = RGB(  0,   0, 128);
  *dp++ = RGB(  0, 128,   0);
  *dp++ = RGB(  0, 128, 128);
  *dp++ = RGB(128,   0,   0);
  *dp++ = RGB(128,   0, 128);
  *dp++ = RGB(128, 128,   0);
  *dp++ = RGB(128, 128, 128);
  *dp++ = RGB(192, 192, 192);
  *dp++ = RGB(  0,   0, 255);
  *dp++ = RGB(  0, 255,   0);
  *dp++ = RGB(  0, 255, 255);
  *dp++ = RGB(255,   0,   0);
  *dp++ = RGB(255,   0, 255);
  *dp++ = RGB(255, 255,   0);
  *dp++ = RGB(255, 255, 255);

  /* Set BYTE ptr p just past ColorTable. */
  p = (LPBYTE) dp;

  /* Outer loop indexes on BYTEs of mp[]; inner loop writes 2 pixels in
   * our color if hi bit set, circulating thru the bits by shifting left.
   */
  for (i = 0; i < 8; i++)
    {
    row = mp[i];
    for (j = 0; j < 4; j++)
      {
      hi = row & 0x80;
      *p = (BYTE) (hi ? color<<4 : 0xF0);
      row <<= 1;
      hi = row & 0x80;
      *p += (hi ? color : 0x0F);
      p++;
      row <<= 1;
      }  /* for j */
    }  /* for i */

  hBrush = CreateDIBPatternBrush(hbmMem, DIB_RGB_COLORS);
  GlobalUnlock(hbmMem);
  GlobalFree(hbmMem);
  return(hBrush);
}

void LOCAL SetPen(HDC hDC, LPBYTE q, int when)
/*
   USE:  Create and select pen.
   IN:   hDC = DC in which to select pen.
         q = ptr to WPG LINEATTR data
         when = (SET_INIT, SET_MID, SET_END) for 1st, middle, or last call
   NOTE: WPG_LINEATTR byte#1   = line style (16 possible)
                      byte#2   = line color
                      byte#3/4 = line width in wpu (WORD)
         Must delete previously created pens. Calling logic must call
         here once at the start with SET_INIT, then possibly multiple times
         with SET_MID (to make settings from WPG_LINEATTR records), then once
         at end with SET_END; all to make sure previously created pens are
         deleted (after selecting a new pen into the DC).
*/
{
  BYTE style;     /* line style (16 possible) */
  BYTE color;     /* line color */
  WORD width;     /* line width (in wpu) */
  HPEN hPen;      /* pen to be created and selected */
  static HPEN hLastPen;   /* last pen selected -- deleted later */

  switch (when)
    {
    case SET_INIT:
      hLastPen = CreatePen(PS_SOLID, 1, RGB(0,0,0));
      SelectObject(hDC, hLastPen);
      break;
    case SET_MID:
      /* Get line style, color, and width from WPG record. */
      style = *q++;
      color = *q;
      width = (WORD)*((LPDWORD) q);
      /* Suppress 'value never used' warning. */
      width = width;
      /* Ignore style (except LA_NOLINE) and width for now ... */
      if (style == LA_NOLINE)
        hPen = CreatePen(PS_NULL, 1, WPGtoWINColor[color]);
      else
        hPen = CreatePen(PS_SOLID, 1, WPGtoWINColor[color]);
      SelectObject(hDC, hPen);
      DeleteObject(hLastPen);
      hLastPen = hPen;
      break;  /* case SET_MID */
    case SET_END:
      SelectObject(hDC, GetStockObject(BLACK_PEN));
      DeleteObject(hLastPen);
      break;
    }  /* switch */
}

void LOCAL DrawLine(HDC hDC, LPBYTE q, SCALE sc)
/*
   USE:  Draw line record.
   IN:   hDC = DC in which to draw
         q = ptr to line data
         sc = SCALE struct to use in drawing line
   NOTE: WPG_LINE record stores (x1, y1, x2, y2) as WORDs.
*/
{
  LPWORD wp;       /* WORD ptr so can dereference data as WORDs */
  WORD   x1, y1;   /* 1st endpoint */
  WORD   x2, y2;   /* 2nd endpoint */

  /* Set wp, as WORD ptr, to line data. */
  wp = (LPWORD)q;

  /* Grab endpoints. */
  x1 = *wp++;
  y1 = *wp++;
  x2 = *wp++;
  y2 = *wp;

  /* Apply scale transformation. */
  x1 = (WORD)(sc.scaleX * x1 + sc.offsX);
  y1 = (WORD)(sc.scaleY * y1 + sc.offsY);
  x2 = (WORD)(sc.scaleX * x2 + sc.offsX);
  y2 = (WORD)(sc.scaleY * y2 + sc.offsY);

  /* Draw line. */
  MoveTo(hDC, x1, y1);
  LineTo(hDC, x2, y2);
}

void LOCAL DrawPoly(HDC hDC, LPBYTE q, SCALE sc, BYTE rt)
/*
   USE:  Draw polygon or polyline record.
   IN:   hDC = DC in which to draw
         q = ptr to polygon data
         sc = SCALE struct to use in drawing polygon
         rt = RecordType (WPG_POLYLINE or WPG_POLYGON)
   NOTE: WPG_POLYLINE and WPG_POLYGON records store number of vertices,
         then the vertices (x1, y1, x2, y2, ..), all as WORDS. Windows'
         Polyline() and CreatePolygonRgn() take a similar format, so
         can use one routine for both.
*/
{
//  WORD    numVertices;  /* number of vertices in polygon */
  int     numVertices;  /* number of vertices in polygon */
  LPPOINT pp;           /* ptr to polygon POINTs */
  int     i;            /* to traverse polygon for scaling */
  HANDLE  hPoly;        /* handle to mem for scaled pts */
  LPPOINT sq;           /* ptr for scaled poly pts (doesn't change) */
  LPPOINT spp;          /* ptr int scaled poly pts (changes) */
  HRGN    hRgn;         /* for PaintRgn() to fill with brush */

  /* Grab numVertices. */
  numVertices = *((LPWORD) q);

  /* Advance q to polygon coordinates. */
  q += sizeof(WORD);

  /* Set pp, as POINT ptr, to polygon coordinates. */
  pp = (LPPOINT)q;

  /* GlobalAlloc() for scaled pts; if can't, return. */
  if (NULL == (hPoly = GlobalAlloc(GHND, numVertices * sizeof(POINT))))
    return;

  /* Generate ptr to store scaled pts. */
  sq = spp = (LPPOINT) GlobalLock(hPoly);

  /* Scale coordinates. */
  for (i = 0; i < numVertices; i++, pp++, spp++)
    {
    spp->x = (int)(sc.scaleX * pp->x + sc.offsX);
    spp->y = (int)(sc.scaleY * pp->y + sc.offsY);
    }

  /* If polygon, first fill the interior. */
  if (rt == WPG_POLYGON)
    {
    hRgn = CreatePolygonRgn(sq, numVertices, ALTERNATE);
    PaintRgn(hDC, hRgn);
    DeleteObject(hRgn);
    }

  /* Draw outline. */
  Polyline(hDC, sq, numVertices);

  /* If polygon, close figure by connecting to start point. */
  if (rt == WPG_POLYGON)
    {
    MoveTo(hDC, sq[numVertices-1].x, sq[numVertices-1].y);
    LineTo(hDC, sq[0].x, sq[0].y);
    }

  /* Done with buffer: unlock and free mem handle. */
  GlobalUnlock(hPoly);
  GlobalFree(hPoly);
}

void LOCAL DrawEllipse(HDC hDC, LPBYTE q, SCALE sc)
/*
   USE:  Draw ellipse record.
   IN:   hDC = DC in which to draw
         q = ptr to ellipse data
         sc = SCALE struct to use in drawing ellipse
   NOTE: WPG_ELLIPSE record stores data all as WORDs: (xCenter, yCenter,
         xRadius, yRadius, rotAngle, startAngle, endAngle, flags).
         Angles in degrees. Starting and ending angle given for ellipse
         in standard orientation, then rotated thru rotation angle.
         Bit#0 of flags set to connect ends to center (pie slice);
         bit#1 of flags set to connect ends to each other (chord).
         Pie slice, chord, rotation not implemented.
*/
{
  LPWORD wp;       /* WORD ptr so can dereference data as WORDs */
  WORD xCenter;    /* x-coord of ellipse center */
  WORD yCenter;    /* y-coord of ellipse center */
  WORD xRadius;    /* half the major axis */
  WORD yRadius;    /* half the minor axis */
  WORD rotAngle;   /* rotation angle (degrees) */
  WORD stAngle;    /* starting angle for arc (degrees) */
  WORD endAngle;   /* ending angle for arc (degrees) */
  WORD flags;      /* for pie slice or chord */
  int  x1, y1;     /* upper left corner of ellipse */
  int  x2, y2;     /* lower right corner of ellipse */

  /* Set wp, as WORD ptr, to ellipse data. */
  wp = (LPWORD)q;

  /* Grab ellipse data. */
  xCenter =  *wp++;
  yCenter =  *wp++;
  xRadius =  *wp++;
  yRadius =  *wp++;
  rotAngle = *wp++;
  stAngle =  *wp++;
  endAngle = *wp++;
  flags    = *wp;

  /* If ellipse to be rotated, return -- will not try it. */
  if (rotAngle) return;

  /* Suppress 'value never used' warning. */
  flags = flags;

  /* Scale center coordinates and radii. */
  xCenter = (WORD)(sc.scaleX * xCenter + sc.offsX);
  yCenter = (WORD)(sc.scaleY * yCenter + sc.offsY);
  xRadius = (WORD) (ABS(sc.scaleX) * xRadius + 0.5);
  yRadius = (WORD) (ABS(sc.scaleY) * yRadius + 0.5);

  /* Create bounding box for Windows' Ellipse(). */
  x1 = xCenter - xRadius;
  y1 = yCenter - yRadius;
  x2 = xCenter + xRadius;
  y2 = yCenter + yRadius;

  /* Draw Arc or Ellipse (angles == 0, 360 indicate full ellipse). */
  if (stAngle != 0 || endAngle != 360)
    {
    stAngle = 360 - stAngle;
    endAngle = 360 - endAngle;
    /* Last four parms in Arc() are coords of two points in indicated
     * directions with respect to the center; 10000 is large enough
     * for good precision without danger of integer overflow.
     */
    Arc(hDC, x1, y1, x2, y2, xCenter + 10000 * cos(DR * stAngle),
                             yCenter + 10000 * sin(DR * stAngle),
                             xCenter + 10000 * cos(DR * endAngle),
                             yCenter + 10000 * sin(DR * endAngle));
    }
  else
    Ellipse(hDC, x1, y1, x2, y2);
}

void LOCAL FrameOurRect(HDC hDC, LPRECT lpr)
/*
   USE: Draw dotted rectangle with little black boxes around it.
   IN:  hDC = DC on which to draw rect
        lpr = ptr to rect to draw
*/
{
  HPEN   hPen;      /* dashed pen */
  HPEN   hOldPen;   /* original pen */
  HBRUSH hOldBrush; /* original brush */
  int    midx;      /* middle x coord of rect */
  int    midy;      /* middle y coord of rect */

  hPen = CreatePen(PS_DOT, 1, 0L);
  hOldPen = SelectObject(hDC, hPen);

  MoveTo(hDC, lpr->left,  lpr->top);
  LineTo(hDC, lpr->right, lpr->top);
  LineTo(hDC, lpr->right, lpr->bottom);
  LineTo(hDC, lpr->left,  lpr->bottom);
  LineTo(hDC, lpr->left,  lpr->top);

  SelectObject(hDC, hOldPen);
  DeleteObject(hPen);

  midx = (lpr->left + lpr->right)/2;
  midy = (lpr->top + lpr->bottom)/2;

  hOldBrush = SelectObject(hDC, GetStockObject(BLACK_BRUSH));

  /* Draw 8 little black boxes starting at top left, proceeding clockwise. */
  Rectangle(hDC, lpr->left - BOXSIZE_PL_MARGIN, lpr->top - BOXSIZE_PL_MARGIN,
                 lpr->left - MARGIN, lpr->top - MARGIN);
  Rectangle(hDC, midx - HALF_BOXSIZE,  lpr->top - BOXSIZE_PL_MARGIN,
                 midx + HALF_BOXSIZE1, lpr->top - MARGIN);
  Rectangle(hDC, lpr->right + MARGIN1, lpr->top - BOXSIZE_PL_MARGIN,
                 lpr->right + BOXSIZE_PL_MARGIN1, lpr->top - MARGIN);
  Rectangle(hDC, lpr->right + MARGIN1, midy - HALF_BOXSIZE,
                 lpr->right + BOXSIZE_PL_MARGIN1, midy + HALF_BOXSIZE1);
  Rectangle(hDC, lpr->right + MARGIN1, lpr->bottom + MARGIN1,
                 lpr->right + BOXSIZE_PL_MARGIN1, lpr->bottom + BOXSIZE_PL_MARGIN1);
  Rectangle(hDC, midx - HALF_BOXSIZE,  lpr->bottom + MARGIN1,
                 midx + HALF_BOXSIZE1, lpr->bottom + BOXSIZE_PL_MARGIN1);
  Rectangle(hDC, lpr->left - BOXSIZE_PL_MARGIN, lpr->bottom + MARGIN1,
                 lpr->left - MARGIN, lpr->bottom + BOXSIZE_PL_MARGIN1);
  Rectangle(hDC, lpr->left - BOXSIZE_PL_MARGIN, midy - HALF_BOXSIZE,
                 lpr->left - MARGIN, midy + HALF_BOXSIZE1);

  SelectObject(hDC, hOldBrush);
}

void LOCAL InitScaleParms(HANDLE hMem, LPSCALE lps, LPRECT lpr)
/*
   USE:  Initialize target rectangle and scale parameters.
   IN:   hMem = handle to memory containing WPG data.
   OUT:  lps = ptr to scale parameters structure
         lpr = ptr to target rectangle : region on the screen into
               which image is mapped.
  NOTE: Uses global ClientRect, sets global WPGsp.
*/
{
  LPBYTE lpWPG;    /* ptr to WPG data */
  double scaleX;   /* X scale factor */
  double scaleY;   /* Y scale factor */
  int    top;      /* top of target rect if loose fit in y */
  int    left;     /* left of target rect if loose fit in x */

  /* Generate ptr to WPG data. */
  lpWPG = (LPBYTE) GlobalLock(hMem);

  /* If find WPG_START1 as first record, use settings there for
   * WPG space; otherwise use defaults set at initialization.
   */
  if (*(lpWPG + WPG_OFFS_FIRSTREC) == WPG_START1)
    WPGsp = *((LPSIZE)(lpWPG + WPG_OFFS_WPGSIZE));

  /* Unlock memory handle. */
  GlobalUnlock(hMem);

  /* Calc scale factors in both dimensions. */
  scaleX = (double) ClientRect.right  / WPGsp.cx;
  scaleY = (double) ClientRect.bottom / WPGsp.cy;

  /* Fit target rectangle into client rectangle with same aspect
   * ratio as WPG space and allowing for margins.
   */
  if (scaleX < scaleY)
    {
    /* TargetRect fits snugly left and right, with extra space
     * top and bottom.
     */
    scaleX *= (1.0 - 2.0 * INIT_MARGIN);
    /* Calc target rect with margin based on INIT_MARGIN on left
     * and right, margin of top units on top and bottom.
     */
    top = (int)((ClientRect.bottom - scaleX * WPGsp.cy) / 2);
    lpr->left   = (int)(ClientRect.right * INIT_MARGIN);
    lpr->top    = top;
    lpr->right  = ClientRect.right * (1.0 - INIT_MARGIN);
    lpr->bottom = ClientRect.bottom - top;
    /* Calculate scale parameters based on this TargetRect. */
    CalcScaleParms(WPGsp, *lpr, lps);
    }
  else
    {
    /* Target rect fits snugly top and bottom, with extra space
     * left and right.
     */
    scaleY *= (1.0 - 2.0 * INIT_MARGIN);
    /* Calc target rect with margin based on INIT_MARGIN on top
     *  and bottom, margin of left units on left and right.
     */
    left = (ClientRect.right - scaleY * WPGsp.cx) / 2;
    lpr->top    = ClientRect.bottom * INIT_MARGIN;
    lpr->left   = left;
    lpr->bottom = ClientRect.bottom * (1.0 - INIT_MARGIN);
    lpr->right  = ClientRect.right - left;
    /* Calculate scale parameters based on this TargetRect. */
    CalcScaleParms(WPGsp, *lpr, lps);
    }
}

int LOCAL PtinHotSpot(LPRECT lpr, POINT pt)
/*
   USE:  Return hot spot index pt is in.
   IN:   lpr = ptr to main rectangle
         pt  = pt to check
   RET:  index of hot spot (TOP_LEFT...TRANSLATE, NO_HIT if not in any)
   NOTE: The hot spots are the main rectangle and 8 little boxes around it.
         NO_HIT is returned if point in none of these hot spots.
*/
{
  int    midx;      /* middle x coord of rect */
  int    midy;      /* middle y coord of rect */

  midx = (lpr->left + lpr->right)/2;
  midy = (lpr->top + lpr->bottom)/2;

  /* If any of these conditions fail, pt is outside expanded bbox. */
  if ((pt.x < (lpr->left  - BOXSIZE_PL_MARGIN)) ||
      (pt.x > (lpr->right + BOXSIZE_PL_MARGIN)) ||
      (pt.y < (lpr->top - BOXSIZE_PL_MARGIN))   ||
      (pt.y > (lpr->bottom + BOXSIZE_PL_MARGIN)))
    return (NO_HIT);

  /* If get thru to here, pt is in expanded bbox. */

  /* Check for pt inside rect. */
  if ((pt.x >= lpr->left) && (pt.x <= lpr->right) &&
      (pt.y >= lpr->top)  && (pt.y <= lpr->bottom))
    return (TRANSLATE);

  /* If first condition met, pt is in horizontal strip along top. */
  if (pt.y <= (lpr->top - MARGIN1))
    {
    /* Check top left box. */
    if (pt.x <= (lpr->left - MARGIN1))
      return (TOP_LEFT);
    /* Check top middle box. */
    if ((pt.x >= (midx - HALF_BOXSIZE)) && (pt.x <= (midx + HALF_BOXSIZE)))
      return (TOP_MID);
    /* Check top right box. */
    if (pt.x >= (lpr->right + MARGIN1))
      return (TOP_RT);
    }

  /* If first condition met, pt is in vertical strip along right. */
  if (pt.x >= (lpr->right + MARGIN1))
    {
    /* Check middle right box. */
    if ((pt.y >= (midy - HALF_BOXSIZE)) && (pt.y <= (midy + HALF_BOXSIZE)))
      return (MID_RT);
    /* Check bottom right box. */
    if (pt.y >= (lpr->bottom + MARGIN1))
      return (BOT_RT);
    }

  /* If first condition met, pt is in horizontal strip along bottom. */
  if (pt.y >= (lpr->bottom + MARGIN1))
    {
    /* Check bottom middle box. */
    if ((pt.x >= (midx - HALF_BOXSIZE)) && (pt.x <= (midx + HALF_BOXSIZE)))
      return (BOT_MID);
    /* Check bottom left box. */
    if (pt.x <= (lpr->left - MARGIN1))
      return (BOT_LEFT);
    }

  /* Only one left to check is middle left box. */
  if (pt.x <= (lpr->left - MARGIN1))
    if ((pt.y >= (midy - HALF_BOXSIZE)) && (pt.y <= (midy + HALF_BOXSIZE)))
      return (MID_LEFT);

  /* If get thru to here, not in any of the hot spots. */
  return (NO_HIT);
}

void LOCAL PlayWPGFile(HANDLE hb, HDC hDC, SCALE sc)
/*
   USE: Play WPG file on DC (render it).
   IN:  hb = handle to memory block containing WPG file
        hDC = DC on which to play the file
        sc = SCALE struct to use in playing file
*/
{
  LPBYTE lpBuf;     /* ptr to buffer */
  LPBYTE cp;        /* current ptr into WPG buffer */
  BYTE   recType;   /* WPG record type */

  /* Lock mem handle to access file data. */
  lpBuf = (LPBYTE) GlobalLock(hb);

  /* Go to 1st record (past WPG prefix). */
  cp = lpBuf + WPG_OFFS_FIRSTREC;

  /* Init SetBrush() and SetPen() -- must be called once at start. */
  SetBrush(hDC, cp, SET_INIT);
  SetPen(hDC, cp, SET_INIT);

  /* Draw records until get out-of-range record or WPG_END record. */
  do 
    recType = DrawWPGRecord(hDC, &cp, sc);
  while (recType >= WPG_FIRST && recType <= WPG_LAST && recType != WPG_END);

  /* End SetBrush() and SetPen() -- must be called once at end. */
  SetBrush(hDC, cp, SET_END);
  SetPen(hDC, cp, SET_END);

  /* Done with buffer: unlock mem handle. */
  GlobalUnlock(hb);
}

void LOCAL CopyWMFToClipBoard(HWND hWnd)
/*
   USE:  Copies WPG file to ClipBoard as MetaFile.
   IN:   hWnd : handle to window
   NOTE: See Petzold, Programming Windows, Chap 16:
         "Following these calls, the hGMem handle (the memory block
         containing the metafile picture structure) and the hmf
         handle (the metafile itself) become invalid for the program
         that created them."
         Routine assumes global memory handle hBuf is valid and contains
         a WPG file.
*/
{
  GLOBALHANDLE   hGMem;   /* global mem for METAFILEPICT */
  LPMETAFILEPICT lpMFP;   /* to point to METAFILEPICT */
  HDC  hdcMeta;           /* handle to MetaFile DC */
  SCALE mScale;           /* SCALE struct for MetaFile */

  hGMem = GlobalAlloc(GHND, (DWORD)sizeof(METAFILEPICT));
  lpMFP = (LPMETAFILEPICT)GlobalLock(hGMem);
  lpMFP->mm   = MM_ANISOTROPIC;
  /* Set MetaFile extents to size of WPG space. */
  lpMFP->xExt = WPGsp.cx;
  lpMFP->yExt = WPGsp.cy;

  /* Generate hDC for memory MetaFile. */
  hdcMeta = CreateMetaFile(NULL);

  /* Set Window Origin and Extent in MetaFile. */
  SetWindowOrg(hdcMeta, 0, 0);
  SetWindowExt(hdcMeta, TargetRect.right - TargetRect.left,
                        TargetRect.bottom - TargetRect.top);

  /* Set SCALE struct with screen's scale factors but xlated to origin. */
  mScale.scaleX = Scale.scaleX;
  mScale.scaleY = Scale.scaleY;
  mScale.offsX  = Scale.offsX - TargetRect.left;
  mScale.offsY  = Scale.offsY - TargetRect.top;
  PlayWPGFile(hBuf, hdcMeta, mScale);

  /* CloseMetaFile(hdcMeta) returns MetaFile handle to send to ClipBoard. */
  lpMFP->hMF = CloseMetaFile(hdcMeta);

  GlobalUnlock(hGMem);
  OpenClipboard(hWnd);
  EmptyClipboard();
  SetClipboardData(CF_METAFILEPICT, hGMem);
  CloseClipboard();
}

void LOCAL PaintInstructionLine(HDC hDC)
/*
   USE:  Paints instruction line at bottom of client area.
   IN:   hDC = DC on which to paint.
*/
{
  char szInstr[50];   /* instruction string */

  /* Draw light gray rectangle at bottom of client rect. */
  SelectObject(hDC, GetStockObject(NULL_PEN));
  SelectObject(hDC, GetStockObject(LTGRAY_BRUSH));
  Rectangle(hDC, 0, ClientRect.bottom - ILINE_HT,
                 ClientRect.right + 1, ClientRect.bottom + 1);
  SelectObject(hDC, GetStockObject(BLACK_PEN));
  /* Draw black line one pixel above light gray rectangle. */
  MoveTo(hDC, 0, ClientRect.bottom - (ILINE_HT+2));
  LineTo(hDC, ClientRect.right, ClientRect.bottom - (ILINE_HT+2));
  /* Choose ANSI_VAR_FONT (MS Sans Serif). */
  SelectObject(hDC, GetStockObject(ANSI_VAR_FONT));
  /* Set TRANSPARENT background mode so gray not over-written. */
  SetBkMode(hDC, TRANSPARENT);
  /* Load instruction string and write it. */
  LoadString(hInst, IDS_INSTR, szInstr, sizeof(szInstr));
  TextOut(hDC, 4, ClientRect.bottom - ILINE_HT, szInstr, lstrlen(szInstr));
}

BOOL LOCAL OpenDlgBox(HWND hWnd, LPSTR szFile, WORD bufLen)
/*
   USE:  Call CommonDlgBox routine GetOpenFileName().
   IN:   hWnd = handle of parent window
         szFile = buffer for user-chosen file name (incl path)
         bufLen = length of szFile buffer
   OUT:  User-chosen file name from DlgBox put in szFile
   RET:  TRUE if ok, FALSE if error or no name chosen (by CANCEL, etc).
   NOTE: See Win31 help on GetOpenFileName() for implementation.
         szFile must be allocated by caller and must be long enough to
         hold the entire path + file name. szDirName holds the default
         path; the first time thru, set to the Windows directory
         ("C:\WINDOWS" or whatever); after that, keep track of last
         user directory and use it the next time.
*/
{
  HINSTANCE hInst;          /* instance handle for this window */
  char szFilter[40];     /* resource str with wildcard specs */
  UINT cbString;         /* num chars in szFilter */
  char chReplace;        /* string separator for szFilter */
  int i;                 /* counter to scan szFilter */
  OPENFILENAME ofn;      /* FileName struct for GetOpenFileName() */
  char szFileTitle[13];  /* user file name (without path) */
  LPSTR p;               /* ptr to last '\' in FileName */

  static BOOL beenThru;        /* says if have been thru function */
  static char szDirName[80];   /* user directory */

  /* The first time thru, get the Windows directory, or the directory
   * stored in SHOWWPG.INI if there is one. Set BOOL beenThru so only
   * come thru here once each time the program runs.
   */
  if (!beenThru)
    {
    /* Get Windows directory name, store in szDirName. */
    GetWindowsDirectory(szDirName, sizeof(szDirName));
    /* Override with last directory from INI file, if there. */
    GetPrivateProfileString("Defaults", "Path", szDirName, szDirName,
                            sizeof(szDirName), "SHOWWPG.INI");
    beenThru = TRUE;
    }

  /* GetClassWord() returns fields set when window was registered:
   * handle to instance, handle to icon, handle to cursor, etc.
   */
  hInst = (HINSTANCE)GetClassWord(hWnd, GCW_HMODULE);
  /* Load resource string in RC with wildcard specs into szFilter. */
  if ((cbString = LoadString(hInst, IDS_FILTERSTRING, szFilter,
       sizeof(szFilter))) == 0)
    return(TRUE);

  /* Get separator char at end of szFilter. */
  chReplace = szFilter[cbString - 1];

  /* Replace separator chars in szFilter with '\0'. */
  for (i = 0; szFilter[i] != '\0'; i++)
    if (szFilter[i] == chReplace)
      szFilter[i] = '\0';

  /* Fill entire structure with 0s. */
  memset(&ofn, 0, sizeof(OPENFILENAME));

  /* szFile is buffer for FileName ('\0' if no init). */
  szFile[0] = '\0';

  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hWnd;
  ofn.lpstrFilter = szFilter;
  ofn.nFilterIndex = 1;
  ofn.lpstrFile = szFile;
  ofn.nMaxFile = bufLen;
  ofn.lpstrFileTitle = szFileTitle;
  ofn.nMaxFileTitle = sizeof(szFileTitle);
  ofn.lpstrInitialDir = szDirName;
  ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;

  /* GetOpenFileName() puts up the Open DlgBox and returns TRUE if ok,
   * FALSE if error, cancel, etc. Pass return value back to caller.
   */
  if (GetOpenFileName(&ofn))
    {
    /* If '\' in FileName, copy to DirName, retaining path only, so
     * user's path will be used next time. _fstrrchr() locates the
     * last '\' in a string, if there is one.
     */
    if ((p = _fstrrchr(ofn.lpstrFile, '\\')) != NULL)
      {
      lstrcpy(szDirName, ofn.lpstrFile);
      *(szDirName + (p - ofn.lpstrFile)) = '\0';
      }
    return(TRUE);  /* ok */
    }
  else
    /* GetOpenFileName() returned error or was cancelled. */
    return(FALSE);
}

void LOCAL RestrictCursor(int type, HWND hWnd, LPPOINT lpp)
/*
   USE:  Restrict cursor depending on type of drag.
   IN:   type = type of drag = TOP_LEFT, TOP_MID ... TRANSLATE
         hWnd = handle to main window
         lpp = ptr to user point
   NOTE: Restrict cursor to rectangle determined by anchor corner
         opposite hot spot and corner of Client rect on the side
         of the hot spot. Settings slightly offset (+1, -2 ..) so
         restrict rectangle, hence TargetRect, never disappear.
         ClipCursor() requires absolute screen coords. Uses
         global ClientRect.
*/
{
  RECT tRectScr;       /* TargetRect in screen coords */
  RECT restrictRect;   /* rectangle to restrict cursor in screen coords */

  /* Copy TargetRect to tRectScr, convert to screen coords. */
  tRectScr = TargetRect;
  ClientToScreen(hWnd, (LPPOINT) &tRectScr);
  ClientToScreen(hWnd, (LPPOINT) &tRectScr + 1);

  /* Copy ClientRect to restrictRect, convert to screen coords. */
  restrictRect = ClientRect;
  ClientToScreen(hWnd, (LPPOINT) &restrictRect);
  ClientToScreen(hWnd, (LPPOINT) &restrictRect + 1);

  /* Raise the restricted area above the instruction line. */
  restrictRect.bottom -= (ILINE_HT + 2);

  /* Reset restrictRect corner or edge opposite hot spot. */
  switch (type)
    {
    case TOP_LEFT:
      restrictRect.right  = tRectScr.right - 1;
      restrictRect.bottom = tRectScr.bottom - 1;
      break;
    case TOP_MID:
      restrictRect.bottom = tRectScr.bottom - 1;
      break;
    case TOP_RT:
      restrictRect.left   = tRectScr.left + 2;
      restrictRect.bottom = tRectScr.bottom - 1;
      break;
    case MID_RT:
      restrictRect.left   = tRectScr.left + 2;
      break;
    case BOT_RT:
      restrictRect.left   = tRectScr.left + 2;
      restrictRect.top    = tRectScr.top + 2;
      break;
    case BOT_MID:
      restrictRect.top    = tRectScr.top + 2;
      break;
    case BOT_LEFT:
      restrictRect.top    = tRectScr.top + 2;
      restrictRect.right  = tRectScr.right - 1;
      break;
    case MID_LEFT:
      restrictRect.right  = tRectScr.right - 1;
      break;
    case TRANSLATE:
      /* Set global deltas -- down pt wrt top left of TargetRect. */
      HSDelta.x = lpp->x - TargetRect.left;
      HSDelta.y = lpp->y - TargetRect.top;
      break;
    }  /* switch */

  /* Restrict cursor. */
  ClipCursor(&restrictRect);
}

void LOCAL DrawDraggingRect(int type, HDC hDC, POINT pt)
/*
   USE:  Draw target rectangle for dragging (without knobs).
   IN:   type = type of drag = TOP_LEFT, TOP_MID ... TRANSLATE
         hDC  = DC to draw on
         pt   = user point to which hot spot has been moved
*/
{
  RECT r;    /* rect to draw, with corners anchor TargetRect and pt */

  /* Init rect to draw as TargetRect. */
  r = TargetRect;

  /* Reset some coords to match user pt. */
  switch (type)
    {
    case TOP_LEFT:
      r.left = pt.x;  r.top  = pt.y;  break;
    case TOP_MID:
      r.top  = pt.y;  break;
    case TOP_RT:
      r.right = pt.x;  r.top = pt.y;  break;
    case MID_RT:
      r.right = pt.x;  break;
    case BOT_RT:
      r.right  = pt.x;  r.bottom = pt.y;  break;
    case BOT_MID:
      r.bottom = pt.y;  break;
    case BOT_LEFT:
      r.left   = pt.x;  r.bottom = pt.y;  break;
    case MID_LEFT:
      r.left = pt.x;  break;
    case TRANSLATE:
      /* Calculate rect based on user pt, HSDelta, and TargetRect. */
      r.left = pt.x - HSDelta.x;
      r.top  = pt.y - HSDelta.y;
      r.right  = r.left + (TargetRect.right - TargetRect.left);
      r.bottom = r.top  + (TargetRect.bottom - TargetRect.top);
      break;
    }  /* switch */

    /* Draw rectangle (just lines, no fill wanted). */
    MoveTo(hDC, r.left, r.top);
    LineTo(hDC, r.right, r.top);
    LineTo(hDC, r.right, r.bottom);
    LineTo(hDC, r.left, r.bottom);
    LineTo(hDC, r.left, r.top);
}

void LOCAL UpdateScale(int type, POINT pt)
/*
   USE:  Update TargetRect and scale parameters depending on type of drag.
   IN:   type = type of drag = TOP_LEFT, TOP_MID ... TRANSLATE
         pt   = final user point determining new parameters
   NOTE: TargetRect updated depending on type of drag; with TOP_LEFT drag,
         only top and left of TargetRect changed, and so on. Then
         CalcScaleParms() called to recalculate scale parameters.
         Also frees cursor, restricted in RestrictCursor().
*/
{
  int dx, dy;   /* offsets in TRANSLATE, increments in ZOOMs */

  /* Update TargetRect. */
  switch (type)
    {
    case TOP_LEFT:
      TargetRect.left = pt.x;  TargetRect.top = pt.y;  break;
    case TOP_MID:
      TargetRect.top = pt.y;  break;
    case TOP_RT:
      TargetRect.right = pt.x;  TargetRect.top = pt.y;  break;
    case MID_RT:
      TargetRect.right = pt.x;  break;
    case BOT_RT:
      TargetRect.right = pt.x;  TargetRect.bottom = pt.y; break;
    case BOT_MID:
      TargetRect.bottom = pt.y;  break;
    case BOT_LEFT:
      TargetRect.left = pt.x;  TargetRect.bottom = pt.y; break;
    case MID_LEFT:
      TargetRect.left = pt.x;  break;
    case TRANSLATE:
      /* Calculate deltas (amount of translation). */
      dx = pt.x - (TargetRect.left + HSDelta.x);
      dy = pt.y - (TargetRect.top  + HSDelta.y);
      /* Offset TargetRect. */
      TargetRect.left += dx;   TargetRect.right  += dx;
      TargetRect.top  += dy;   TargetRect.bottom += dy;
      break;
    case ZOOM_IN:
      /* Calculate deltas (amount of expansion). */
      dx = SCALE_PERCENT * (TargetRect.right - TargetRect.left);
      dy = SCALE_PERCENT * (TargetRect.bottom - TargetRect.top);
      /* Expand TargetRect. */
      TargetRect.left -= dx;   TargetRect.right  += dx;
      TargetRect.top  -= dy;   TargetRect.bottom += dy;
      break;
    case ZOOM_OUT:
      /* Calculate deltas (amount of contraction). */
      dx = SCALE_PERCENT * (TargetRect.right - TargetRect.left);
      dy = SCALE_PERCENT * (TargetRect.bottom - TargetRect.top);
      /* Contract TargetRect. */
      TargetRect.left += dx;   TargetRect.right  -= dx;
      TargetRect.top  += dy;   TargetRect.bottom -= dy;
      break;
    }  /* switch */

  /* Recalculate scale parameters based on new TargetRect. */
  CalcScaleParms(WPGsp, TargetRect, &Scale);

  /* No longer restrict cursor. */
  ClipCursor(NULL);
}

void LOCAL CalcScaleParms(SIZE wpgsp, RECT tr, LPSCALE lps)
/*
   USE:  Calculate scale parameters from WPG space size and TargetRect.
   IN:   wpgsp = WPG space size, x = width, y = height
         tr    = target rect on screen where image goes
   OUT:  lps = ptr to struct with scale parameters
*/
{
  /* Scale equation solutions in x. */
  lps->scaleX = (double) (tr.right - tr.left) / wpgsp.cx;
  lps->offsX  = tr.left;

  /* Scale equation solutions in y. */
  lps->scaleY = (double) (tr.top - tr.bottom) / wpgsp.cy;
  lps->offsY  = tr.bottom;
}
